<!doctype html>
<html lang="en">
  <link
    rel="stylesheet"
    href="https://cdn.jsdelivr.net/npm/@picocss/pico@2/css/pico.min.css"
  />
  <style>
    .container {
      max-width: 40rem;
    }
  </style>
  <main class="container">{% block content %}{{ content }}{% endblock %}</main>
</html>
