<!doctype html>
<link
  rel="stylesheet"
  href="https://cdn.jsdelivr.net/npm/@picocss/pico@2/css/pico.min.css"
/>
<main class="container">
  <h1>Error: {{ name }}</h1>
  <p>{{ description }}</p>
</main>
