{% extends "_base.html" %} {% block content %}
<h1>OpenID Connect Provider Mock</h1>
<ul>
  <li>
    <a href="https://github.com/geigerzaehler/oidc-provider-mock"
      >Project on Github</a
    >
  </li>
  <li>
    <a href="/.well-known/openid-configuration">OpenID Discovery Document</a>
  </li>
  <li>
    <a href="/oauth2/authorize">Client Authorization</a>
  </li>
  <li>
    <a href="/oidc/login">Self login</a>
  </li>
</ul>
{% endblock %}
